#!/usr/bin/python
# -*- coding: UTF-8 -*-
"""
@author:andrew
@file:Hybrid_Net014.py
@email:admin@marques22.com
@email:2021022362@m.scnu.edu.cn
@time:2022/10/14
"""

import os

import torch
from torch import nn
from torchinfo import summary
from torch import cat

os.environ["CUDA_VISIBLE_DEVICES"] = "0"

# 修改激活函数
# 提高呼吸采样率

# 输入时长
WHOLE_SEGMENT_SECOND = 30

# 呼吸采样率
RESPIRATORY_FRE = 10

# BCG 时频图大小
BCG_GRAPH_SIZE = (26, 121)


class HYBRIDNET021(nn.Module):
    def __init__(self, num_classes=2, init_weights=True):
        super(HYBRIDNET021, self).__init__()

        self.lstm = nn.LSTM(input_size=1,
                            hidden_size=32,
                            num_layers=2,
                            bidirectional=True,
                            batch_first=True)

        self.classifier = nn.Sequential(
            # nn.Dropout(p=0.5),
            nn.Linear(64, 8),
            nn.GELU(),
            nn.Linear(8, num_classes),
        )

        if init_weights:
            self.initialize_weights()

    def initialize_weights(self):
        for m in self.modules():
            if isinstance(m, (nn.Conv2d, nn.Conv1d)):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')  # 何教授方法
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)  # 正态分布赋值
                nn.init.constant_(m.bias, 0)

    def forward(self, x1):
        x1, (_, _) = self.lstm(x1)
        # print(x1.shape)
        x1 = x1[:, -1]

        x1 = torch.flatten(x1, start_dim=1)
        # print(x1.shape)
        # x2 = x2.squeeze()
        # x = torch.cat((x1, x2), dim=1)
        x = x1
        x = self.classifier(x)
        return x


if __name__ == '__main__':
    model = HYBRIDNET021().cuda()
    summary(model, [(32, 300, 1), (32, 3, 1)])
